"""Grid Parameters with instructions for generating SensorGrids."""
from pydantic import Field, constr
from typing import Union, List
from enum import Enum

from honeybee_schema._base import NoExtraBaseModel
from honeybee_schema.altnumber import Autocalculate


class _GridParameterBase(NoExtraBaseModel):
    """Base object for all GridParameters."""

    dimension: float = Field(
        ...,
        gt=0,
        description='The dimension of the grid cells as a number.'
    )

    include_mesh: bool = Field(
        True,
        description='A boolean to note whether the resulting SensorGrid should '
        'include the mesh.'
    )


class RoomGridParameter(_GridParameterBase):
    """Instructions for a SensorGrid generated from a Room2D's floors."""

    type: constr(regex='^RoomGridParameter$') = 'RoomGridParameter'

    offset: float = Field(
        1.0,
        description='A number for how far to offset the grid from the Room2D floors. '
        '(Default: 1.0, suitable for Models in Meters).'
    )

    wall_offset: float = Field(
        0,
        gt=0,
        description='A number for the distance at which sensors close to walls '
        'should be removed. Note that this option has no effect unless the '
        'value is more than half of the dimension.'
    )


class RoomRadialGridParameter(RoomGridParameter):
    """Instructions for a SensorGrid of radial directions around positions from floors.

    This type of sensor grid is particularly helpful for studies of multiple
    view directions, such as imageless glare studies.
    """

    type: constr(regex='^RoomRadialGridParameter$') = 'RoomRadialGridParameter'

    offset: float = Field(
        1.2,
        description='A number for how far to offset the grid from the Room2D floors. '
        '(Default: 1.2, suitable for Models in Meters).'
    )

    dir_count: int = Field(
        8,
        gt=0,
        description='A positive integer for the number of radial directions '
        'to be generated around each position.'
    )

    start_vector: List[float] = Field(
        None,
        description='A vector as 3 (x, y, z) values to set the start direction of '
        'the generated directions. This can be used to orient the resulting sensors to '
        'specific parts of the scene. It can also change the elevation of the '
        'resulting directions since this start vector will always be rotated in '
        'the XY plane to generate the resulting directions.',
        min_items=3,
        max_items=3
    )

    mesh_radius: Union[Autocalculate, float] = Field(
        Autocalculate(),
        ge=0,
        description='An optional number to override the radius of the meshes '
        'generated around each sensor. If Autocalculate, it will be '
        'equal to 45 percent of the grid dimension.'
    )


class ExteriorFaceType(str, Enum):

    wall = 'Wall'
    roof = 'Roof'
    floor = 'Floor'
    all = 'All'


class ExteriorFaceGridParameter(_GridParameterBase):
    """Instructions for a SensorGrid generated from exterior Faces."""

    type: constr(regex='^ExteriorFaceGridParameter$') = 'ExteriorFaceGridParameter'

    offset: float = Field(
        0.1,
        description='A number for how far to offset the grid from the Faces. '
        '(Default: 0.1, suitable for Models in Meters).'
    )

    face_type: ExteriorFaceType = Field(
        ExteriorFaceType.wall,
        description='Text to specify the type of face that will be used to '
        'generate grids. Note that only Faces with Outdoors boundary '
        'conditions will be used, meaning that most Floors will typically '
        'be excluded unless they represent the underside of a cantilever.'
    )

    punched_geometry: bool = Field(
        False,
        description='A boolean to note whether the punched_geometry of the faces '
        'should be used (True) with the areas of sub-faces removed from the grid '
        'or the full geometry should be used (False).'
    )


class ExteriorApertureType(str, Enum):

    window = 'Window'
    skylight = 'Skylight'
    all = 'All'


class ExteriorApertureGridParameter(_GridParameterBase):
    """Instructions for a SensorGrid generated from exterior Aperture."""

    type: constr(regex='^ExteriorApertureGridParameter$') = \
        'ExteriorApertureGridParameter'

    offset: float = Field(
        0.1,
        description='A number for how far to offset the grid from the Apertures. '
        '(Default: 0.1, suitable for Models in Meters).'
    )

    aperture_type: ExteriorApertureType = Field(
        ExteriorApertureType.all,
        description='Text to specify the type of Aperture that will be used to '
        'generate grids. Window indicates Apertures in Walls. Skylights '
        'are in parent Roof faces.'
    )



"""Window Parameters with instructions for generating windows."""
from pydantic import Field, constr, conlist
from typing import Union, List

from honeybee_schema._base import NoExtraBaseModel
from honeybee_schema.altnumber import Autocalculate


class GriddedSkylightArea(NoExtraBaseModel):
    """Gridded skylights defined by an absolute area."""

    type: constr(regex='^GriddedSkylightArea$') = 'GriddedSkylightArea'

    skylight_area: float = Field(
        ...,
        gt=0,
        description='A number for the skylight area in current model units. '
        'If this area is larger than the area of the roof that it is applied '
        'to, the skylight will fill the parent roof at a 99 percent ratio.'
    )

    spacing: Union[Autocalculate, float] = Field(
        Autocalculate(),
        gt=0,
        description='A number for the spacing between the centers of each grid cell. '
        'This should be less than a third of the dimension of the Roof geometry '
        'if multiple, evenly-spaced skylights are desired. If Autocalculate, a spacing '
        'of one third the smaller dimension of the parent Roof will be automatically '
        'assumed.'
    )


class GriddedSkylightRatio(NoExtraBaseModel):
    """Gridded skylights derived from an area ratio with the roof."""

    type: constr(regex='^GriddedSkylightRatio$') = 'GriddedSkylightRatio'

    skylight_ratio: float = Field(
        ...,
        gt=0,
        lt=1,
        description='A number between 0 and 1 for the ratio between the skylight '
        'area and the total Roof face area.'
    )

    spacing: Union[Autocalculate, float] = Field(
        Autocalculate(),
        gt=0,
        description='A number for the spacing between the centers of each grid cell. '
        'This should be less than a third of the dimension of the Roof geometry '
        'if multiple, evenly-spaced skylights are desired. If Autocalculate, a spacing '
        'of one third the smaller dimension of the parent Roof will be automatically '
        'assumed.'
    )


class DetailedSkylights(NoExtraBaseModel):
    """Several detailed skylights defined by 2D Polygons (lists of 2D vertices)."""

    type: constr(regex='^DetailedSkylights$') = 'DetailedSkylights'

    polygons: List[
        conlist(conlist(float, min_items=2, max_items=2), min_items=3)
    ] = Field(
        ...,
        description='An array of arrays with each sub-array representing a polygonal '
        'boundary of a skylight. Each sub-array should consist of arrays '
        'representing points, which contain 2 values for 2D coordinates in '
        'the world XY system. These coordinate values should lie within the '
        'parent Room2D Polygon.'
    )

    are_doors: List[bool] = Field(
        default=None,
        description='An array of booleans that align with the polygons and note '
        'whether each of the polygons represents an overhead door (True) or a '
        'skylight (False). If None, it will be assumed that all polygons represent '
        'skylights and they will be translated to Apertures in any resulting '
        'Honeybee model.'
    )



"""Model radiance properties."""
from pydantic import Field, constr
from typing import List, Union

from honeybee_schema._base import NoExtraBaseModel
from honeybee_schema.radiance.modifier import _REFERENCE_UNION_MODIFIERS
from honeybee_schema.radiance.modifierset import ModifierSet, ModifierSetAbridged
from honeybee_schema.radiance.global_modifierset import GlobalModifierSet

from .gridpar import RoomGridParameter, RoomRadialGridParameter, \
    ExteriorFaceGridParameter, ExteriorApertureGridParameter


class Room2DRadiancePropertiesAbridged(NoExtraBaseModel):

    type: constr(regex='^Room2DRadiancePropertiesAbridged$') = \
        'Room2DRadiancePropertiesAbridged'

    modifier_set: str = Field(
        default=None,
        description='Identifier of a ModifierSet to specify all modifiers for '
        'the Room2D. If None, the Room2D will use the Story or Building '
        'modifier_set or the Model global_modifier_set. Any ModifierSet '
        'assigned here will override those assigned to the parent objects.'
    )

    grid_parameters: List[
        Union[RoomGridParameter, RoomRadialGridParameter,
              ExteriorFaceGridParameter, ExteriorApertureGridParameter]
    ] = Field(
        default=None,
        description='An optional list of GridParameter objects to describe '
        'how sensor grids should be generated for the Room2D.'
    )


class StoryRadiancePropertiesAbridged(NoExtraBaseModel):

    type: constr(regex='^StoryRadiancePropertiesAbridged$') = \
        'StoryRadiancePropertiesAbridged'

    modifier_set: str = Field(
        default=None,
        description='Name of a ModifierSet to specify all modifiers for '
        'the Story. If None, the Story will use the Building modifier_set '
        'or the Model global_modifier_set. Any ModifierSet '
        'assigned here will override those assigned to the parent objects.'
    )


class BuildingRadiancePropertiesAbridged(NoExtraBaseModel):

    type: constr(regex='^BuildingRadiancePropertiesAbridged$') = \
        'BuildingRadiancePropertiesAbridged'

    modifier_set: str = Field(
        default=None,
        description='Name of a ModifierSet to specify all modifiers for '
        'the Building. If None, the Model global_modifier_set will be used.'
    )


class ContextShadeRadiancePropertiesAbridged(NoExtraBaseModel):

    type: constr(regex='^ContextShadeRadiancePropertiesAbridged$') = \
        'ContextShadeRadiancePropertiesAbridged'

    modifier: str = Field(
        default=None,
        description='Name of a Modifier to set the reflectance and '
        'specularity of the ContextShade. If None, the the default '
        'of 0.2 diffuse reflectance will be used.'
    )


class ModelRadianceProperties(NoExtraBaseModel):

    type: constr(regex='^ModelRadianceProperties$') = 'ModelRadianceProperties'

    global_modifier_set: GlobalModifierSet = Field(
        default=GlobalModifierSet(),
        description='Global Radiance modifier set.',
        readOnly=True
    )

    modifier_sets: List[Union[ModifierSetAbridged, ModifierSet]] = Field(
        default=None,
        description='List of all ModifierSets in the Model.'
    )

    modifiers: List[_REFERENCE_UNION_MODIFIERS] = Field(
        default=None,
        description='A list of all unique modifiers in the model. This includes '
        'modifiers across all the Model modifier_sets.'
    )
